@@ -181,7 +181,7 @@ tlsg_getfile( const char *path, gnutls_datum_t *buf )
  * initialize a new TLS context
  */
 static int
-tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
+tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server, char *errmsg )
 {
 	tlsg_ctx *ctx = lo->ldo_tls_ctx;
 	int rc;
@@ -195,20 +195,26 @@ tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
  	}
 
 	if (lo->ldo_tls_cacertdir != NULL) {
-		rc = gnutls_certificate_set_x509_trust_dir(
-			ctx->cred,
-			lt->lt_cacertdir,
-			GNUTLS_X509_FMT_PEM );
-		if ( rc > 0 ) {
-			Debug2( LDAP_DEBUG_TRACE,
-				"TLS: loaded %d CA certificates from directory `%s'.\n",
-				rc, lt->lt_cacertdir );
-		} else {
-			Debug1( LDAP_DEBUG_ANY,
-				"TLS: warning: no certificate found in CA certificate directory `%s'.\n",
-				lt->lt_cacertdir );
-			/* only warn, no return */
+		char **dirs = ldap_str2charray( lt->lt_cacertdir, CERTPATHSEP );
+		int i;
+		for ( i=0; dirs[i]; i++ ) {
+			rc = gnutls_certificate_set_x509_trust_dir(
+				ctx->cred,
+				dirs[i],
+				GNUTLS_X509_FMT_PEM );
+			if ( rc > 0 ) {
+				Debug2( LDAP_DEBUG_TRACE,
+					"TLS: loaded %d CA certificates from directory `%s'.\n",
+					rc, dirs[i] );
+			} else {
+				Debug1( LDAP_DEBUG_ANY,
+					"TLS: warning: no certificate found in CA certificate directory `%s'.\n",
+					dirs[i] );
+				/* only warn, no return */
+				strncpy( errmsg, gnutls_strerror( rc ), ERRBUFSIZE );
+			}
 		}
+		ldap_charray_free( dirs );
 	}
 
 	if (lo->ldo_tls_cacertfile != NULL) {
@@ -228,6 +234,7 @@ tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 				"TLS: warning: no certificate loaded from CA certificate file `%s'.\n",
 				lo->ldo_tls_cacertfile );
 			/* only warn, no return */
+			strncpy( errmsg, gnutls_strerror( rc ), ERRBUFSIZE );
 		}
 	}
 
@@ -244,6 +251,7 @@ tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 				"TLS: could not use CA certificate: %s (%d)\n",
 				gnutls_strerror( rc ),
 				rc );
+			strncpy( errmsg, gnutls_strerror( rc ), ERRBUFSIZE );
 			return -1;
 		}
 	}
@@ -256,7 +264,10 @@ tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 		unsigned int max = VERIFY_DEPTH;
 
 		rc = gnutls_x509_privkey_init( &key );
-		if ( rc ) return -1;
+		if ( rc ) {
+			strncpy( errmsg, gnutls_strerror( rc ), ERRBUFSIZE );
+			return -1;
+		}
 
 		/* OpenSSL builds the cert chain for us, but GnuTLS
 		 * expects it to be present in the certfile. If it's
@@ -285,6 +296,7 @@ tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 				"TLS: could not use private key: %s (%d)\n",
 				gnutls_strerror( rc ),
 				rc );
+			strncpy( errmsg, gnutls_strerror( rc ), ERRBUFSIZE );
 			return rc;
 		}
 
@@ -310,6 +322,7 @@ tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 				"TLS: could not use certificate: %s (%d)\n",
 				gnutls_strerror( rc ),
 				rc );
+			strncpy( errmsg, gnutls_strerror( rc ), ERRBUFSIZE );
 			return rc;
 		}
 
@@ -333,6 +346,7 @@ tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 				"TLS: could not use certificate with key: %s (%d)\n",
 				gnutls_strerror( rc ),
 				rc );
+			strncpy( errmsg, gnutls_strerror( rc ), ERRBUFSIZE );
 			return -1;
 		}
 	} else if (( lo->ldo_tls_certfile || lo->ldo_tls_keyfile )) {
@@ -350,7 +364,10 @@ tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 			ctx->cred,
 			lt->lt_crlfile,
 			GNUTLS_X509_FMT_PEM );
-		if ( rc < 0 ) return -1;
+		if ( rc < 0 ) {
+			strncpy( errmsg, gnutls_strerror( rc ), ERRBUFSIZE );
+			return -1;
+		}
 		rc = 0;
 	}
 
@@ -369,7 +386,10 @@ tlsg_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 			rc = gnutls_dh_params_import_pkcs3( ctx->dh_params, &buf,
 				GNUTLS_X509_FMT_PEM );
 		LDAP_FREE( buf.data );
-		if ( rc ) return -1;
+		if ( rc ) {
+			strncpy( errmsg, gnutls_strerror( rc ), ERRBUFSIZE );
+			return -1;
+		}
 		gnutls_certificate_set_dh_params( ctx->cred, ctx->dh_params );
 	}
 
@@ -945,13 +965,12 @@ tlsg_session_pinning( LDAP *ld, tls_session *sess, char *hashalg, struct berval
 	}
 
 	if ( hashalg ) {
-		len = gnutls_hash_get_len( alg );
-		keyhash.bv_val = LDAP_MALLOC( len );
+		keyhash.bv_len = gnutls_hash_get_len( alg );
+		keyhash.bv_val = LDAP_MALLOC( keyhash.bv_len );
 		if ( !keyhash.bv_val || gnutls_fingerprint( alg, &key,
-					keyhash.bv_val, &len ) < 0 ) {
+					keyhash.bv_val, &keyhash.bv_len ) < 0 ) {
 			goto done;
 		}
-		keyhash.bv_len = len;
 	} else {
 		keyhash.bv_val = (char *)key.data;
 		keyhash.bv_len = key.size;
